#!/usr/bin/python3

import argparse
from encoder import *
import vettingMethod as vm
import igraph as ig
import json
from pathlib import Path
from subprocess import Popen, PIPE, STDOUT, run
import os

class testFunctions:
	def testQuteNoLazy(graph):
		enc = basicEncoding(graph, lazyOptimization=False)
		p = Popen(["./qute"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
		p.communicate(bytes(str(enc), encoding="utf-8"))
		return p.returncode == 10
	
	def testCqestoNoLazy(graph):
		enc = basicEncoding(graph, lazyOptimization=False)
		p = Popen(["./cqesto"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
		p.communicate(bytes(str(enc), encoding="utf-8"))
		return p.returncode == 10

	def testEncoderNoLazyOptimization(graph):
		enc = basicEncoding(graph, lazyOptimization=False)
		p = Popen(["./cqesto"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
		p.communicate(bytes(str(enc), encoding="utf-8"))
		return p.returncode == 10
	
	def testEncoderNoLazyNoOverlappingOptimization(graph):
		enc = basicEncoding(graph, lazyOptimization=False, overlappingOptimization=False)
		p = Popen(["./cqesto"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
		p.communicate(bytes(str(enc), encoding="utf-8"))
		return p.returncode == 10

	def testEncoderNoOverlappingOptimization(graph):
		enc = basicEncoding(graph, overlappingOptimization=False)
		p = Popen(["./cqesto"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
		p.communicate(bytes(str(enc), encoding="utf-8"))
		return p.returncode == 10

	def testEncoderNoLazyOptimizationAllColored(graph):
		enc = basicEncoding(graph, lazyOptimizationAllColored=False)
		p = Popen(["./cqesto"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
		p.communicate(bytes(str(enc), encoding="utf-8"))
		return p.returncode == 10

	def testEncoderNoLazyOptimizationOverlapping(graph):
		enc = basicEncoding(graph, lazyOptimizationOverlapping=False)
		p = Popen(["./cqesto"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
		p.communicate(bytes(str(enc), encoding="utf-8"))
		return p.returncode == 10

	def testEncoderNoExactlyOneOptimization(graph):
		enc = basicEncoding(graph, exactlyOneOptimization=False)
		p = Popen(["./cqesto"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
		p.communicate(bytes(str(enc), encoding="utf-8"))
		return p.returncode == 10

	def testEncoderPure(graph):
		enc = basicEncoding(graph)
		p = Popen(["./cqesto"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
		p.communicate(bytes(str(enc), encoding="utf-8"))
		return p.returncode == 10

	def testEncoderSubsets(graph):
		nonReduicbleSubsets = vm.calcNonReduicbleSubsets(graph)
		enc = encodingOnlyColorabilityClasses(graph, nonReduicbleSubsets)
		p = Popen(["./cqesto"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
		p.communicate(bytes(str(enc), encoding="utf-8"))
		return p.returncode == 10
	
	def testEncoderQuteSubsets(graph):
		nonReduicbleSubsets = vm.calcNonReduicbleSubsets(graph)
		enc = encodingOnlyColorabilityClasses(graph, nonReduicbleSubsets)
		p = Popen(["./qute"], stdout=PIPE, stdin=PIPE, stderr=PIPE)
		p.communicate(bytes(str(enc), encoding="utf-8"))
		return p.returncode == 10

	def testVettingMethod(graph):
		return vm.testIfChoosable(graph)

def parentRoutine(file, timeout, suiteId):
	suiteList = [
		[
			"testEncoderPure",
			"testEncoderSubsets",
			"testEncoderQuteSubsets",
			"testVettingMethod"
		],
		[	
			"testEncoderPure",
			"testEncoderNoLazyOptimization",
			"testEncoderNoOverlappingOptimization",
			"testEncoderNoLazyOptimizationAllColored",
			"testEncoderNoLazyOptimizationOverlapping",
			"testEncoderNoExactlyOneOptimization",
		],
		[
			"testQuteNoLazy",
			"testCqestoNoLazy"
		],
		[
			"testEncoderNoLazyOptimization",
			"testEncoderNoLazyNoOverlappingOptimization"
		]
	]
	suite = suiteList[suiteId]


	timeoutArgs = []
	if timeout:
		timeoutArgs = ["timeout", timeout]

	for testConfig in suite:
		process = run(["sudo", "chrt", "-f", "99"] + timeoutArgs + ["/usr/bin/time", "--quiet", "-f", '{"elapsed":"%E"}', "./benchmarks.py", "-f" , testConfig, "-i", file], capture_output=True)
		res = {}
		
		try:
			print(process.stderr.decode())
			print(process.stdout.decode())
			res["elapsed"] = json.loads(process.stderr.decode())["elapsed"]
		except json.decoder.JSONDecodeError:
			res["elapsed"] = None
		if process.returncode != 124:
			res["terminated"] = True
			res["truthValue"] = process.returncode == 10
		else:
			res["terminated"] = False

		Path(file.replace(".json", f"{testConfig}.json")).write_text(json.dumps(res))

def generatorRoutine(dir, timeout, suiteId):
	configFile = json.loads(Path(f"{dir}/config.json").read_text())


	if configFile["mode"] == "randomConstantDegreeGraphs":
		config = configFile["config"]
		for session in config:
			for graphIndex in range(0, session["occurances"]):
				for reducedVertex in range(-1, session["numVertices"]):
					print(f"working on: test_n{session['numVertices']}_deg{session['vertexDegrees']}_i{graphIndex}_rv{reducedVertex}")
					graph = ig.Graph.K_Regular(session["numVertices"], session["vertexDegrees"])
			
					listSizeFunction = [session["startingConstantListSizeFunction"]] * graph.vcount()

					if reducedVertex >=0:
						listSizeFunction[reducedVertex] -=1

					content = {
						"graph": list(map(lambda x: x.tuple, graph.es)),
						"listSizeFunction": listSizeFunction,
					}

					testDir = f"{dir}/graphs/test_n{session['numVertices']}_deg{session['vertexDegrees']}_i{graphIndex}_rv{reducedVertex}"
					if not os.path.exists(testDir):
						os.makedirs(testDir)

					testPath = f"{testDir}/test.json"

					Path(testPath).write_text(json.dumps(content))

					parentRoutine(testPath, timeout, suiteId)

	elif configFile["mode"] == "handpicked":
		tests = configFile["tests"]
		for instance in tests:
			graph = ig.Graph(instance["graph"])
			listSizeFunction = instance["listSizeFunction"]
			print(f"testing graph:")
			print(graph)
			print(f"with list size function {listSizeFunction}")
			
			content = {
				"graph": list(map(lambda x: x.tuple, graph.es)),
				"listSizeFunction": listSizeFunction,
			}
			
			testDir = f"{dir}/graphsHandpicked/withRecursiveCalls_{instance['recursiveCalls']}"
			if not os.path.exists(testDir):
				os.makedirs(testDir)

			testPath = f"{testDir}/testGraph.json"

			Path(testPath).write_text(json.dumps(content))

			parentRoutine(testPath, timeout, suiteId)




parser = argparse.ArgumentParser()

parser.add_argument("-g", action="store_true", help="generate graphs from config")
parser.add_argument("-p", action="store_true", help="run as parent")
parser.add_argument("-t", nargs=1, required=False, help="timeout for each process")
parser.add_argument("-s", nargs=1, required=False, help="specify the test suite to run")
parser.add_argument("-f", nargs=1, required=False, help="specify the function to test")
parser.add_argument("-i", nargs=1, help="specify the input file")

args = parser.parse_args()

suite = 0
if args.s:
	suite = int(args.s[0])
if args.g:
	generatorRoutine(args.i[0], args.t[0], suite)
	exit()

if args.p:
	parentRoutine(args.i[0], args.t[0], suite)
	exit()

readResult = json.loads(Path(args.i[0]).read_text())

graph = ig.Graph(readResult["graph"])
graph.vs[:]["listsize"] = readResult["listSizeFunction"]
graph.vs[:]["originalIndex"] = range(0,graph.vcount())

if getattr(testFunctions, args.f[0])(graph):
	exit(10)
exit(20)