#!/usr/bin/python3
from typing import List
from qcir import *
import igraph as ig


#graph must be preprocessed to be correct, maxColorability classes can be set, if there is a lower bound than n-1
#returns a tuple of list variables and color variables
def variables(graph: ig.Graph):
	if not "maxColorabilityClasses" in graph.attributes():
		graph["maxColorabilityClasses"] = graph.vcount() - 1

	graph.vs["listAssignment"] = [["x_{}_{}".format(i, j) for j in range(graph["maxColorabilityClasses"])] for i in range(graph.vcount())]
	graph.vs["colorAssignment"] = [["c_{}_{}".format(i, j) for j in range(graph["maxColorabilityClasses"])] for i in range(graph.vcount())]

#no check if listAssignments and color Assignments have the same lengths
def inLists(graph: ig.Graph):
	listAssignments = graph.vs[:]["listAssignment"]
	colorAssignments = graph.vs[:]["colorAssignment"]

	helperGatesOut = [["inListsHelper_{}_{}".format(i, j) for j in range(len(listAssignments[i]))] for i in range(len(listAssignments))]
	helperGates = [Gate(helperGatesOut[i][j], GateConnective.OR, [Literal(True, listAssignments[i][j]), Literal(False, colorAssignments[i][j])]) for i in range((len(listAssignments))) for j in range(len(listAssignments[i]))]
	inListsGate = Gate("inLists", GateConnective.AND, [Literal(True, helperGatesOut[i][j]) for i in range(len(listAssignments)) for j in range(len(listAssignments[i]))])

	allGates = helperGates
	allGates.append(inListsGate)

	return ("inLists", allGates)

def properColoring(graph: ig.Graph):
	helperGatesOut = []
	helperGates = []
	properColoringLiterals = []

	for (v1, v2) in graph.get_edgelist():
		for i in range(graph["maxColorabilityClasses"]):
			colorAssignment1 = graph.vs[v1]["colorAssignment"]
			colorAssignment2 = graph.vs[v2]["colorAssignment"]
			helperGateOut = "properColoring_v{}_v{}_c{}".format(v1, v2,i)

			helperGatesOut.append(helperGateOut)
			helperGates.append(Gate(helperGateOut, GateConnective.OR, [Literal(False, colorAssignment1[i]), Literal(False, colorAssignment2[i])]))
			properColoringLiterals.append(Literal(True, helperGateOut))

	properColoringOut = Gate("properColoring", GateConnective.AND, properColoringLiterals)

	allGates = helperGates
	allGates.append(properColoringOut)

	return ("properColoring", allGates)

def allColored(graph: ig.Graph, lazyOptimization=True, exactlyOne=False):
	helperGatesOut = []
	helperGates = []
	allColoredLiterals = []
	allColoredOverflowVertexOut = ["allColoredOverflow_{}".format(i) for i in range(len(graph.vs()))]

	for i,vertex in enumerate(graph.vs()):
		colorList = graph.vs[i]["colorAssignment"]
		helperGateOut = "allColored_{}".format(i)

		helperGatesOut.append(helperGateOut)
		lazyLiterals = []
		if lazyOptimization:
			lazyLiterals = [Literal(True, vertex["lazyChoosable"])]
		helperGates.append(Gate(helperGateOut, GateConnective.OR, lazyLiterals + [Literal(True, var) for var in colorList]))
		allColoredLiterals.append(Literal(True, helperGateOut))

		if exactlyOne:
			atLeastOneUntilOut = ["atLeastOneUntil_{}_{}".format(i,j) for j in range(graph["maxColorabilityClasses"])]
			allColoredOverflowsOut = ["allColoredOverflow_{}_{}".format(i, j) for j in range(graph["maxColorabilityClasses"])]

			firstAtLeastGate = Gate(atLeastOneUntilOut[0], GateConnective.OR, [Literal(True, colorList[0])])
			helperGates.append(firstAtLeastGate)

			otherAtLeastGates = [Gate(atLeastOneUntilOut[j], GateConnective.OR, [Literal(True, colorList[j]), Literal(True, atLeastOneUntilOut[j-1])]) for j in range(1,graph["maxColorabilityClasses"])]
			helperGates += otherAtLeastGates

			allColoredOverflowGates = [Gate(allColoredOverflowsOut[j], GateConnective.AND, [Literal(True, colorList[j]), Literal(True, atLeastOneUntilOut[j-1])]) for j in range(1,graph["maxColorabilityClasses"])]
			helperGates += allColoredOverflowGates

			allColoredOverflowGate = Gate(allColoredOverflowVertexOut[i], GateConnective.OR, [Literal(True, overflowOut) for overflowOut in allColoredOverflowsOut[1:]])
			helperGates.append(allColoredOverflowGate)


	
	allGates = helperGates
	if not exactlyOne:
		allGates.append(Gate("allColored", GateConnective.AND, allColoredLiterals))
	if exactlyOne:
		allGates.append(Gate("allColored", GateConnective.AND, [Literal(False, acvo) for acvo in allColoredOverflowVertexOut] + allColoredLiterals))

	return("allColored", allGates)

#would not work if there was a list of size 1 (those get pruned during preprocessing)
def fullListAssginment(graph: ig.Graph):
	#here, all helper gates are collected
	allHelperGates = []
	vertexListFullLiterals = []
	
	#these are the literals that ensure that the respective vertex is colored correctly
	vertexLiterals = []

	for i,vertex in enumerate(graph.vs, 0):

		localHelpers = [["fullHelper_{}_{}_{}".format(i,j,k) for k in range(vertex["listsize"])] for j in range(len(vertex["listAssignment"]))]
		#the fullCountUpHelper at pos 0 will never be used
		countUpHelpers = [["fullCountUpHelper_{}_{}_{}".format(i,j,k) for k in range(vertex["listsize"])] for j in range(len(vertex["listAssignment"]))]
		overflowDetectors = ["overflowDetector_{}_{}".format(i,j) for j in range(len(vertex["listAssignment"]))]

		helperGateOut = "fullVertex_{}".format(i)
		overflowGateOut = "overflowGate_{}".format(i)
		vertexLiterals.append(Literal(True, helperGateOut))

		baseGate = Gate(localHelpers[0][0] , GateConnective.AND, [Literal(True, vertex["listAssignment"][0])])
		allHelperGates.append(baseGate)

		firstRowGates = [Gate(localHelpers[j][0], GateConnective.OR, [Literal(True, vertex["listAssignment"][j]), Literal(True, localHelpers[j-1][0])]) for j in range(1,len(vertex["listAssignment"]))]
		allHelperGates = allHelperGates + firstRowGates

		#these gates are set to false, since every helper but the first one is set to false
		firstColGates = [Gate(localHelpers[0][j], GateConnective.AND, [Literal(True, localHelpers[0][0]), Literal(False, localHelpers[0][0])]) for j in range(1,vertex["listsize"])]
		allHelperGates = allHelperGates + firstColGates

		countUpGates = [[Gate(countUpHelpers[j][k], GateConnective.AND, [Literal(True, localHelpers[j-1][k-1]), Literal(True, vertex["listAssignment"][j])]) for k in range(1, vertex["listsize"])] for j in range(1, len(vertex["listAssignment"]))]
		localGates = [[Gate(localHelpers[j][k], GateConnective.OR, [Literal(True, countUpHelpers[j][k]), Literal(True, localHelpers[j-1][k])]) for k in range(1, vertex["listsize"])] for j in range(1, len(vertex["listAssignment"]))]

		for countUpI in range(0, len(vertex["listAssignment"])-1):
			allHelperGates = allHelperGates + countUpGates[countUpI]
			allHelperGates = allHelperGates + localGates[countUpI]

		overflowHelperGates = [Gate(overflowDetectors[j], GateConnective.AND, [Literal(True, localHelpers[j-1][len(localHelpers[j-1])-1]),Literal(True, vertex["listAssignment"][j])]) for j in range(1,len(vertex["listAssignment"]))]
		allHelperGates = allHelperGates + overflowHelperGates
		overflowGate = Gate(overflowGateOut, GateConnective.OR, [Literal(True, overflowHelperGate.variable) for overflowHelperGate in overflowHelperGates])
		allHelperGates.append(overflowGate)

		localHelpersLastI = len(localHelpers)-1
		atLeastKOut = localHelpers[localHelpersLastI][len(localHelpers[localHelpersLastI]) - 1]
		helperGate = Gate(helperGateOut, GateConnective.AND, [Literal(True, atLeastKOut), Literal(False, overflowGateOut)])
		allHelperGates.append(helperGate)

		vertexListFullLiterals.append(Literal(True, helperGateOut))

	allGates = allHelperGates
	allGates.append(Gate("fullListAssignment", GateConnective.AND, vertexListFullLiterals))
	return("fullListAssignment", allGates)

def colorClassesOrdered(graph: ig.Graph, increment = 1):
	allHelperGates = []

	#note that the indexing is still vertex,color class, even though in this case, the color class takes precedence
	#ordered means that the less than relation does not hold
	orderedBiggerClass = [f"orderedBiggerClass_{i}_{increment}" for i in range(graph["maxColorabilityClasses"] - increment)]
	orderedBiggerHelpers = [[f"orderedBigger_{i}_{j}_{increment}" for j in range(graph["maxColorabilityClasses"] - increment)] for i in range(graph.vcount() + 1)]
	orderedBiggerLocalHelpers = [[f"orderedBiggerLocal_{i}_{j}_{increment}" for j in range(graph["maxColorabilityClasses"] - increment)] for i in range(graph.vcount())]
	orderedSmallerLocalHelpers = [[f"orderedSmallerLocal_{i}_{j}_{increment}" for j in range(graph["maxColorabilityClasses"] - increment)] for i in range(graph.vcount())]
	orderedBiggerUntilHelpers = [[f"orderedBiggerUntil_{i}_{j}_{increment}" for j in range(graph["maxColorabilityClasses"] - increment)] for i in range(graph.vcount())]
	orderedSmallerUntilHelpers = [[f"orderedSmallerUntil_{i}_{j}_{increment}" for j in range(graph["maxColorabilityClasses"] - increment)] for i in range(graph.vcount())]

	orderedBiggerLocalGates = [[Gate(orderedBiggerLocalHelpers[i][j], GateConnective.AND, [Literal(True, graph.vs["listAssignment"][i][j]), Literal(False, graph.vs["listAssignment"][i][j+1])]) for j in range(graph["maxColorabilityClasses"] - increment)] for i in range(graph.vcount())]
	for gates in orderedBiggerLocalGates:
		allHelperGates = allHelperGates + gates

	orderedBiggerUntilFirstGates = [Gate(orderedBiggerUntilHelpers[0][i], GateConnective.OR, [Literal(True, orderedBiggerLocalHelpers[0][i])]) for i in range(graph["maxColorabilityClasses"] - increment)]
	allHelperGates = allHelperGates + orderedBiggerUntilFirstGates

	orderedBiggerUntilGates = [[Gate(orderedBiggerUntilHelpers[i][j], GateConnective.OR, [Literal(True, orderedBiggerUntilHelpers[i-1][j]), Literal(True, orderedBiggerLocalHelpers[i][j])]) for j in range(graph["maxColorabilityClasses"] - increment)] for i in range(1, graph.vcount())]
	for gates in orderedBiggerUntilGates:
		allHelperGates = allHelperGates + gates

	
	orderedSmallerLocalGates = [[Gate(orderedSmallerLocalHelpers[i][j], GateConnective.AND, [Literal(False, graph.vs["listAssignment"][i][j]), Literal(True, graph.vs["listAssignment"][i][j+1])]) for j in range(graph["maxColorabilityClasses"] - increment)] for i in range(graph.vcount())]
	for gates in orderedSmallerLocalGates:
		allHelperGates = allHelperGates + gates

	orderedSmallerUntilFirstGates = [Gate(orderedSmallerUntilHelpers[0][i], GateConnective.OR, [Literal(True, orderedSmallerLocalHelpers[0][i])]) for i in range(graph["maxColorabilityClasses"] - increment)]
	allHelperGates = allHelperGates + orderedSmallerUntilFirstGates

	orderedSmallerUntilGates = [[Gate(orderedSmallerUntilHelpers[i][j], GateConnective.OR, [Literal(True, orderedSmallerUntilHelpers[i-1][j]), Literal(True, orderedSmallerLocalHelpers[i][j])]) for j in range(graph["maxColorabilityClasses"] - increment)] for i in range(1, graph.vcount())]
	for gates in orderedSmallerUntilGates:
		allHelperGates = allHelperGates + gates


	orderedBiggerFirstGates = [Gate(orderedBiggerHelpers[0][i], GateConnective.AND, [Literal(True, orderedBiggerUntilHelpers[0][i])]) for i in range(graph["maxColorabilityClasses"] - increment)]
	allHelperGates = allHelperGates + orderedBiggerFirstGates

	orderedBiggerGates = [[Gate(orderedBiggerHelpers[i][j], GateConnective.AND, [Literal(False, orderedSmallerUntilHelpers[i-1][j]), Literal(True, orderedBiggerUntilHelpers[i][j])]) for j in range(graph["maxColorabilityClasses"] - increment)] for i in range(1, graph.vcount())]
	for gates in orderedBiggerGates:
		allHelperGates = allHelperGates + gates

	#the last gate is necessary to handle the equals case
	orderedBiggerLastGates = [Gate(orderedBiggerHelpers[graph.vcount()][i], GateConnective.AND, [Literal(False, orderedSmallerUntilHelpers[graph.vcount() - 2][i]), Literal(False, orderedSmallerUntilHelpers[graph.vcount() - 1][i])]) for i in range(graph["maxColorabilityClasses"] - increment)]
	allHelperGates = allHelperGates + orderedBiggerLastGates


	biggerClassGates = [Gate(orderedBiggerClass[i], GateConnective.OR, [Literal(True, orderedBiggerHelpers[j][i]) for j in range(graph.vcount() + 1)]) for i in range(graph["maxColorabilityClasses"] - increment)]
	allHelperGates = allHelperGates + biggerClassGates

	out = f"colorClassesOrderd_{increment}"
	allGates = allHelperGates
	allGates.append(Gate(out, GateConnective.AND, [Literal(True, obc) for obc in orderedBiggerClass]))
	return(out, allGates)

# extremely bad perfomance, maybe some bug?
def colorClassesOrderedIfThenElse(graph: ig.Graph):
	allHelperGates = []

	#note that the indexing is still vertex,color class, even though in this case, the color class takes precedence
	#ordered means that the less than relation does not hold
	orderedBiggerClass = [f"orderedBiggerClass_{i}" for i in range(graph["maxColorabilityClasses"] - 1)]
	orderedBiggerHelpers = [[f"orderedBigger_{i}_{j}" for j in range(graph["maxColorabilityClasses"])] for i in range(graph.vcount() + 1)]
	orderedBiggerLocalHelpers = [[f"orderedBiggerLocal_{i}_{j}" for j in range(graph["maxColorabilityClasses"])] for i in range(graph.vcount())]
	orderedSmallerLocalHelpers = [[f"orderedSmallerLocal_{i}_{j}" for j in range(graph["maxColorabilityClasses"])] for i in range(graph.vcount())]

	orderedBiggerLocalGates = [[Gate(orderedBiggerLocalHelpers[i][j], GateConnective.AND, [Literal(True, graph.vs["listAssignment"][i][j]), Literal(False, graph.vs["listAssignment"][i][j+1])]) for j in range(graph["maxColorabilityClasses"] - 1)] for i in range(graph.vcount())]
	for gates in orderedBiggerLocalGates:
		allHelperGates = allHelperGates + gates

	orderedSmallerLocalGates = [[Gate(orderedSmallerLocalHelpers[i][j], GateConnective.AND, [Literal(False, graph.vs["listAssignment"][i][j]), Literal(True, graph.vs["listAssignment"][i][j+1])]) for j in range(graph["maxColorabilityClasses"] - 1)] for i in range(graph.vcount())]
	for gates in orderedSmallerLocalGates:
		allHelperGates = allHelperGates + gates

	orderedBiggerLastHelperGates = [Gate(orderedBiggerHelpers[graph.vcount() - 1][j], GateConnective.AND, [Literal(False, orderedSmallerLocalHelpers[graph.vcount() - 1][j])]) for j in range(graph["maxColorabilityClasses"] - 1)]
	allHelperGates += orderedBiggerLastHelperGates
	orderedBiggerHelperGates = [[Gate(orderedBiggerHelpers[i][j], GateConnective.ITE, [Literal(True, orderedBiggerLocalHelpers[i+1][j]), Literal(False, orderedSmallerLocalHelpers[i][j]), Literal(True, orderedBiggerHelpers[i+1][j])]) for j in range(graph["maxColorabilityClasses"] - 1)] for i in reversed(range(graph.vcount() - 1))]
	for gates in orderedBiggerHelperGates:
		allHelperGates += gates

	biggerClassGates = [Gate(orderedBiggerClass[i], GateConnective.OR, [Literal(True, orderedBiggerLocalHelpers[i][0]), Literal(True, orderedBiggerHelpers[i][0])]) for i in range(graph["maxColorabilityClasses"] - 1)]
	allHelperGates = allHelperGates + biggerClassGates

	out = "colorClassesOrderd"
	allGates = allHelperGates
	allGates.append(Gate(out, GateConnective.AND, [Literal(True, obc) for obc in orderedBiggerClass]))
	return(out, allGates)

def colorabilityClassOverlapping(graph: ig.Graph, lazyOptimization=True):
	allHelperGates = []

	colorabilityClassRelevantOut = [["colorabilityClassRelevant_lazy_{}_{}_{}".format(lazyOptimization, i,j) for j in range(graph["maxColorabilityClasses"])] for i in range(graph.vcount())]
	colorabilityClassExistsOut = ["colorabilityClassExists_lazy_{}_{}".format(lazyOptimization, i) for i in range(graph["maxColorabilityClasses"])]
	colorabilityClassLocalOverlapOut = [[["colorabilityClassOverlaps_lazy_{}_{}_{}_{}".format(lazyOptimization, i,j,k) for k in range(graph["maxColorabilityClasses"])] for j in range(graph["maxColorabilityClasses"])] for i in range(graph.vcount())]
	colorabilityClassOverlapsMatrixOut = [["colorabilityClassOverlapsMatrix_lazy_{}_{}_{}".format(lazyOptimization, i,j) for j in range(graph["maxColorabilityClasses"])] for i in range(graph["maxColorabilityClasses"])]
	colorabilityClassOverlapsAllOtherOut = ["colorabilityClassOverlapsAllOther_lazy_{}_{}".format(lazyOptimization, i) for i in range(graph["maxColorabilityClasses"])]

	if lazyOptimization:
		colorabilityClassRelevantGates = [[Gate(colorabilityClassRelevantOut[i][j], GateConnective.AND, [Literal(True, graph.vs["listAssignment"][i][j]), Literal(False, graph.vs[i]["lazyChoosable"])]) for j in range(graph["maxColorabilityClasses"])] for i in range(graph.vcount())]
		for relevantGates in colorabilityClassRelevantGates:
			allHelperGates = allHelperGates + relevantGates
		
		colorabilityClassExistsGates = [Gate(colorabilityClassExistsOut[i], GateConnective.OR, [Literal(True, colorabilityClassRelevantOut[j][i]) for j in range(len(graph.vs["listAssignment"]))]) for i in range(graph["maxColorabilityClasses"])]
		#This requires every colorability class to exist. Therefore, one has to run the solver multiple times until unsat is found or we know that a bad list assignment is found
		#colorabilityClassExistsGates = [Gate(colorabilityClassExistsOut[i], GateConnective.OR, [Literal(True, colorabilityClassRelevantOut[i][0]), Literal(False, colorabilityClassRelevantOut[i][0])]) for i in range(graph["maxColorabilityClasses"])]
		allHelperGates = allHelperGates + colorabilityClassExistsGates

	else:
		colorabilityClassExistsGates = [Gate(colorabilityClassExistsOut[i], GateConnective.OR, [Literal(True, graph.vs["listAssignment"][j][i]) for j in range(len(graph.vs["listAssignment"]))]) for i in range(graph["maxColorabilityClasses"])]
		allHelperGates = allHelperGates + colorabilityClassExistsGates


	mcc = graph["maxColorabilityClasses"]
	la = graph.vs["listAssignment"]
	ccloo = colorabilityClassLocalOverlapOut
	if lazyOptimization:
		colorabilityClassLocalOverlapGates = [[[Gate(ccloo[i][j][k], GateConnective.AND, [Literal(True, colorabilityClassRelevantOut[i][j]), Literal(True, colorabilityClassRelevantOut[i][k])]) for k in range(mcc)] for j in range(mcc)] for i in range(graph.vcount())]
	else:
		colorabilityClassLocalOverlapGates = [[[Gate(ccloo[i][j][k], GateConnective.AND, [Literal(True, la[i][j]), Literal(True, la[i][k])]) for k in range(mcc)] for j in range(mcc)] for i in range(graph.vcount())]
	for gatesList in colorabilityClassLocalOverlapGates:
		for gates in gatesList:
			allHelperGates = allHelperGates + gates
	
	ccomo = colorabilityClassOverlapsMatrixOut
	cceo = colorabilityClassExistsOut
	#if the other class has no vertices, overlap is true
	colorabilityClassOverlapsMatrixGates = [[Gate(ccomo[i][j], GateConnective.OR, [Literal(False, cceo[i]), Literal(False, cceo[j])] + [Literal(True, ccloo[k][i][j]) for k in range(graph.vcount())]) for i in range(mcc)] for j in range(mcc)]
	for gates in colorabilityClassOverlapsMatrixGates:
		allHelperGates = allHelperGates + gates

	ccoaoo = colorabilityClassOverlapsAllOtherOut
	colorabilityClassOverlapsAllOtherGates = [Gate(ccoaoo[i], GateConnective.AND, [Literal(True, ccomo[i][j]) for j in range(mcc)]) for i in range(mcc)]
	allHelperGates = allHelperGates + colorabilityClassOverlapsAllOtherGates

	out = "colorabilityClassOverlapping_lazy_{}".format(lazyOptimization)
	allGates = allHelperGates
	#this version was used before, it performs slightly worse during an informal test
	#allGates.append(Gate(out, GateConnective.AND, [Literal(True, ccOverlaps) for ccOverlaps in ccoaoo]))
	allGates.append(Gate(out, GateConnective.AND, [Literal(True, ccOverlaps) for ccOverlapsLine in ccomo for ccOverlaps in ccOverlapsLine]))
	return(out, allGates)

#if a vertex can be safely colored afterwards, we call it lazy choosable
#correct version: two cases: a) There is a color that no neighbor has b) Less than n neighbors have the color in the list -> on this level, one could let it ripple to overlapping
#incorrect version two cases: a) There is a color in the list no neighbor is colored in b) Less than n neighbors are colored in one of the colors of the list -> here, there might be a vertex which doesn't get a color but receives a lazy colorable "get out of jail" card
#if such a vertex is found, it does not need to be colored any more (since it can be lazily colored anyways) and does not need to be considered for the overlapping Classes optimization
#furthermore, a vertex becomming lazy choosable can have ripple effects on other vertices
#since no circular references are possible, there might be some benefit in having more than one lazy choosable gate for every vertex
#only this version can be used for overlapping

#for ease of operations, we consider only the case for one pass and color appears at no neighbor
def lazyChoosable(graph: ig.Graph, passes = 3):
	allHelperGates = []

	lazyChoosableOut = ["lazyChoosable_{}".format(i) for i in range(graph.vcount())]

	specificColorNotInNeighborListOut = [["specificColorNotInNeighborList_{}_{}".format(i,j) for j in range(graph["maxColorabilityClasses"])] for i in range(graph.vcount())]
	colorNotInNeighborListOut = ["colorNotInNeighborList_{}".format(i) for i in range(graph.vcount())]
	

	specificColorLazyChoosableOut = [[["specificColorLazyChoosable_{}_{}_{}".format(i,j,k) for k in range(passes)] for j in range(graph["maxColorabilityClasses"])] for i in range(graph.vcount())]
	specificColorNotInNeighborListPassOut = [[["specificColorNotInNeighborListPass_{}_{}_{}".format(i,j,k) for k in range(passes)] for j in range(graph["maxColorabilityClasses"])] for i in range(graph.vcount())]
	colorNotInNeighborListPassOut = [["colorNotInNeighborList_{}_{}".format(i, j) for j in range(passes)] for i in range(graph.vcount())]

	listLongEnoughPassOut = [["listLongEnough_{}_{}".format(i,j)for j in range(passes)] for i in range(graph.vcount())]

	lazyChoosablePassOut = [["lazyChoosablePass_{}_{}".format(i, j) for j in range(passes)] for i in range(graph.vcount())]
	

	#here: some color not in any neighbor
	for i, vertex in enumerate(graph.vs):
		specificColorGates = [Gate(specificColorNotInNeighborListOut[i][j], GateConnective.AND, [Literal(True, graph.vs["listAssignment"][i][j])] + [Literal(False, graph.vs[neighbor]["listAssignment"][j]) for neighbor in graph.neighbors(vertex)]) for j in range(graph["maxColorabilityClasses"])]
		allHelperGates = allHelperGates + specificColorGates

		colorNotInNeighborListGate = Gate(colorNotInNeighborListOut[i], GateConnective.OR, [Literal(True, specificColorOut) for specificColorOut in specificColorNotInNeighborListOut[i]])
		allHelperGates.append(colorNotInNeighborListGate)

	#here: color in vertex overlaps less than listsize + 1 neighbors
	for i, vertex in enumerate(graph.vs):
		vertexOverlappingColorListPass0Out = [["vertexOverlappingColorListPass_{}_{}_{}_{}".format(i,j,k,0) for k in range(graph["maxColorabilityClasses"])] for j in graph.neighbors(vertex)]
		vertexOverlappingNeighborPass0Out =  ["vertexOverlappingNeighborPass_{}_{}_{}".format(i,j,0) for j in graph.neighbors(vertex)]
		vertexOverlappingCounterPass0Out = [["vertexOverlappingCounterPass_{}_{}_{}_{}".format(i,j,0,k) for k in range(vertex["listsize"])] for j in graph.neighbors(vertex)]
		vertexOverlappingCountUpHelperPass0Out = [["vertexOverlappingCountUpHelperPass_{}_{}_{}_{}".format(i,j,0,k) for k in range(vertex["listsize"])] for j in graph.neighbors(vertex)]

		vertexOverlappingColorListPass0Gates = [[Gate(vertexOverlappingColorListPass0Out[neighborIndex][color], GateConnective.AND, [Literal(True, vertex["listAssignment"][color]), Literal(True, graph.vs[neighbor]["listAssignment"][color])]) for color in range(graph["maxColorabilityClasses"])] for neighborIndex,neighbor in enumerate(graph.neighbors(vertex))]
		for gates in vertexOverlappingColorListPass0Gates:
			allHelperGates = allHelperGates + gates

		vertexOverlappingNeighborPass0Gates = [Gate(vertexOverlappingNeighborPass0Out[neighborIndex], GateConnective.OR, [Literal(True, vertexOverlappingColorListPass0Out[neighborIndex][color]) for color in range(graph["maxColorabilityClasses"])]) for neighborIndex,neighbor in enumerate(graph.neighbors(vertex))]
		allHelperGates = allHelperGates + vertexOverlappingNeighborPass0Gates

		baseGate = Gate(vertexOverlappingCounterPass0Out[0][0] , GateConnective.AND, [Literal(True, vertexOverlappingNeighborPass0Out[0])])
		allHelperGates.append(baseGate)

		firstRowGates = [Gate(vertexOverlappingCounterPass0Out[j][0], GateConnective.OR, [Literal(True, vertexOverlappingNeighborPass0Out[j]), Literal(True, vertexOverlappingCounterPass0Out[j-1][0])]) for j in range(1,len(graph.neighbors(vertex)))]
		allHelperGates = allHelperGates + firstRowGates
		
		firstColGates = [Gate(vertexOverlappingCounterPass0Out[0][j], GateConnective.AND, [Literal(True, vertexOverlappingCounterPass0Out[0][0]), Literal(False, vertexOverlappingCounterPass0Out[0][0])]) for j in range(1,vertex["listsize"])]
		allHelperGates = allHelperGates + firstColGates

		countUpGates = [[Gate(vertexOverlappingCountUpHelperPass0Out[j][k], GateConnective.AND, [Literal(True, vertexOverlappingCounterPass0Out[j-1][k-1]), Literal(True, vertexOverlappingNeighborPass0Out[j])]) for k in range(1, vertex["listsize"])] for j in range(1, len(graph.neighbors(vertex)))]
		localGates = [[Gate(vertexOverlappingCounterPass0Out[j][k], GateConnective.OR, [Literal(True, vertexOverlappingCountUpHelperPass0Out[j][k]), Literal(True, vertexOverlappingCounterPass0Out[j-1][k])]) for k in range(1, vertex["listsize"])] for j in range(1, len(graph.neighbors(vertex)))]

		for countUpI in range(0, len(graph.neighbors(vertex))-1):
			allHelperGates = allHelperGates + countUpGates[countUpI]
			allHelperGates = allHelperGates + localGates[countUpI]

		listLongEnoughGate = Gate(listLongEnoughPassOut[i][0], GateConnective.AND, [Literal(False, vertexOverlappingCounterPass0Out[len(graph.neighbors(vertex))-1][vertex["listsize"]-1])])
		allHelperGates.append(listLongEnoughGate)


	for i, vertex in enumerate(graph.vs):
		lazyGate = Gate(lazyChoosablePassOut[i][0], GateConnective.OR, [Literal(True, colorNotInNeighborListOut[i]), Literal(True, listLongEnoughPassOut[i][0])])
		allHelperGates.append(lazyGate)

	for lazyPass in range(1, passes):
		#here: some color not in any neighbor
		for i, vertex in enumerate(graph.vs):
			specificColorLazyChoosableGates = [Gate(specificColorLazyChoosableOut[i][j][lazyPass], GateConnective.AND, [Literal(True, graph.vs["listAssignment"][i][j]), Literal(False, lazyChoosablePassOut[i][lazyPass-1])]) for j in range(graph["maxColorabilityClasses"])]
			allHelperGates = allHelperGates + specificColorLazyChoosableGates
		for i, vertex in enumerate(graph.vs):
			specificColorGates = [Gate(specificColorNotInNeighborListPassOut[i][j][lazyPass], GateConnective.AND, [Literal(True, graph.vs["listAssignment"][i][j])] + [Literal(False, specificColorLazyChoosableOut[neighbor][j][lazyPass]) for neighbor in graph.neighbors(vertex)]) for j in range(graph["maxColorabilityClasses"])]
			allHelperGates = allHelperGates + specificColorGates

			colorNotInNeighborListGate = Gate(colorNotInNeighborListPassOut[i][lazyPass], GateConnective.OR, [Literal(True, specificColorOut[lazyPass]) for specificColorOut in specificColorNotInNeighborListPassOut[i]])
			allHelperGates.append(colorNotInNeighborListGate)

		
		#here: color in vertex overlaps less than listsize + 1 neighbors
		for i, vertex in enumerate(graph.vs):
			vertexOverlappingColorListPassOut = [["vertexOverlappingColorListPass_{}_{}_{}_{}".format(i,j,k,lazyPass) for k in range(graph["maxColorabilityClasses"])] for j in graph.neighbors(vertex)]
			vertexOverlappingNeighborPassOut =  ["vertexOverlappingNeighborPass_{}_{}_{}".format(i,j,lazyPass) for j in graph.neighbors(vertex)]
			vertexOverlappingCounterPassOut = [["vertexOverlappingCounterPass_{}_{}_{}_{}".format(i,j,lazyPass,k) for k in range(vertex["listsize"])] for j in graph.neighbors(vertex)]
			vertexOverlappingCountUpHelperPassOut = [["vertexOverlappingCountUpHelperPass_{}_{}_{}_{}".format(i,j,lazyPass,k) for k in range(vertex["listsize"])] for j in graph.neighbors(vertex)]

			vertexOverlappingColorListPassGates = [[Gate(vertexOverlappingColorListPassOut[neighborIndex][color], GateConnective.AND, [Literal(True, vertex["listAssignment"][color]), Literal(False, lazyChoosablePassOut[neighbor][lazyPass-1]), Literal(True, graph.vs[neighbor]["listAssignment"][color])]) for color in range(graph["maxColorabilityClasses"])] for neighborIndex,neighbor in enumerate(graph.neighbors(vertex))]
			for gates in vertexOverlappingColorListPassGates:
				allHelperGates = allHelperGates + gates

			vertexOverlappingNeighborPassGates = [Gate(vertexOverlappingNeighborPassOut[neighborIndex], GateConnective.OR, [Literal(True, vertexOverlappingColorListPassOut[neighborIndex][color]) for color in range(graph["maxColorabilityClasses"])]) for neighborIndex,neighbor in enumerate(graph.neighbors(vertex))]
			allHelperGates = allHelperGates + vertexOverlappingNeighborPassGates

			baseGate = Gate(vertexOverlappingCounterPassOut[0][0] , GateConnective.AND, [Literal(True, vertexOverlappingNeighborPassOut[0])])
			allHelperGates.append(baseGate)

			firstRowGates = [Gate(vertexOverlappingCounterPassOut[j][0], GateConnective.OR, [Literal(True, vertexOverlappingNeighborPassOut[j]), Literal(True, vertexOverlappingCounterPassOut[j-1][0])]) for j in range(1,len(graph.neighbors(vertex)))]
			allHelperGates = allHelperGates + firstRowGates

			firstColGates = [Gate(vertexOverlappingCounterPassOut[0][j], GateConnective.AND, [Literal(True, vertexOverlappingCounterPassOut[0][0]), Literal(False, vertexOverlappingCounterPassOut[0][0])]) for j in range(1,vertex["listsize"])]
			allHelperGates = allHelperGates + firstColGates

			countUpGates = [[Gate(vertexOverlappingCountUpHelperPassOut[j][k], GateConnective.AND, [Literal(True, vertexOverlappingCounterPassOut[j-1][k-1]), Literal(True, vertexOverlappingNeighborPassOut[j])]) for k in range(1, vertex["listsize"])] for j in range(1, len(graph.neighbors(vertex)))]
			localGates = [[Gate(vertexOverlappingCounterPassOut[j][k], GateConnective.OR, [Literal(True, vertexOverlappingCountUpHelperPassOut[j][k]), Literal(True, vertexOverlappingCounterPassOut[j-1][k])]) for k in range(1, vertex["listsize"])] for j in range(1, len(graph.neighbors(vertex)))]

			for countUpI in range(0, len(graph.neighbors(vertex))-1):
				allHelperGates = allHelperGates + countUpGates[countUpI]
				allHelperGates = allHelperGates + localGates[countUpI]

			listLongEnoughGate = Gate(listLongEnoughPassOut[i][lazyPass], GateConnective.AND, [Literal(False, vertexOverlappingCounterPassOut[len(graph.neighbors(vertex))-1][vertex["listsize"] - 1])])
			allHelperGates.append(listLongEnoughGate)
		

		for i, vertex in enumerate(graph.vs):
			lazyGate = Gate(lazyChoosablePassOut[i][lazyPass], GateConnective.OR, [Literal(True, colorNotInNeighborListPassOut[i][lazyPass]), Literal(True, listLongEnoughPassOut[i][lazyPass])])
			#lazyGate = Gate(lazyChoosablePassOut[i][lazyPass], GateConnective.OR, [Literal(True, colorNotInNeighborListPassOut[i][lazyPass])])
			allHelperGates.append(lazyGate)
	

	graph.vs["lazyChoosable"] = [passesOut[passes-1] for passesOut in lazyChoosablePassOut]
	graph.vs["lazyChoosableAllPasses"] = lazyChoosablePassOut
	return allHelperGates

#encodes that every coloability class is element of the given list
def allColorabilityClassesPartOfSubsets(graph: ig.Graph, subsets: List[List]):
	out = "allColorabilityClassesPartOfSubsets"
	colorabilityClassMatchesSubset = [[f"colorabilityClassMatchesSubset_{j}_{k}" for k in range(len(subsets))] for j in range(graph["maxColorabilityClasses"])]
	colorabilityClassMatchesAnySubset = [f"colorabilityClassMatchesAnySubset_{j}" for j in range(graph["maxColorabilityClasses"])]


	allGates = []
	colorabilityClassMatchesSubsetGates = [[Gate(colorabilityClassMatchesSubset[j][k], GateConnective.AND, [Literal(i in subsets[k], graph.vs[i]["listAssignment"][j]) for i in range(graph.vcount())]) for k in range(len(subsets))]for j in range(graph["maxColorabilityClasses"])]
	for gates in colorabilityClassMatchesSubsetGates:
		allGates += gates
	colorabilityClassMatchesAnySubsetGates = [Gate(colorabilityClassMatchesAnySubset[j], GateConnective.OR, [Literal(True, colorabilityClassMatchesSubset[j][k]) for k in range(len(subsets))]) for j in range(graph["maxColorabilityClasses"])]
	allGates += colorabilityClassMatchesAnySubsetGates

	allGates.append(Gate(out, GateConnective.AND, [Literal(True, ccmas) for ccmas in colorabilityClassMatchesAnySubset]))
	return (out, allGates)	


def encodingOnlyColorabilityClasses(graph: ig.Graph, subsets: List[List],
		overlappingOptimization=True,
		exactlyOneOptimization=True):
	try:
		listsizes = graph.vs[:]["listsize"]
	except KeyError:
		print("[Error]: Graph needs listsize attribute, encoding stopped")

	variables(graph)

	(inListsOut, inListsGates) = inLists(graph)
	(properColoringOut, properColoringGates) = properColoring(graph)
	(allColoredOut, allColoredGates) = allColored(graph, lazyOptimization=False, exactlyOne=exactlyOneOptimization)
	(fullListAssignmentOut, fullListAssginmentGates) = fullListAssginment(graph)
	colorClassesOrderedOutList = []
	colorClassesOrderedGatesList = []
	#for i in range(1, graph["maxColorabilityClasses"]):
	for i in range(1, 2):
		(colorClassesOrderedOut, colorClassesOrderedGates) = colorClassesOrdered(graph, i)
		colorClassesOrderedOutList.append(colorClassesOrderedOut)
		colorClassesOrderedGatesList.append(colorClassesOrderedGates)


	#(colorClassesOrderedOut, colorClassesOrderedGates) = colorClassesOrderedIfThenElse(graph)
	(colorabilityClassOverlappingOut, colorabilityClassOverlappingGates) = colorabilityClassOverlapping(graph, lazyOptimization=False)
	(subsetsOut, subsetsGates) = allColorabilityClassesPartOfSubsets(graph, subsets)

	admissible = Gate("admissible", GateConnective.AND, [Literal(True, inListsOut), Literal(True, properColoringOut), Literal(True, allColoredOut)])
	if overlappingOptimization:
		outputGate = Gate("choosable", GateConnective.OR, [Literal(False, ccoo) for ccoo in colorClassesOrderedOutList] + [Literal(False, colorabilityClassOverlappingOut), Literal(False, fullListAssignmentOut), Literal(False, subsetsOut), Literal(True, "admissible")])
	else:
		outputGate = Gate("choosable", GateConnective.OR, [Literal(False, ccoo) for ccoo in colorClassesOrderedOutList] + [Literal(False, fullListAssignmentOut), Literal(False, subsetsOut), Literal(True, "admissible")])
	gateList = inListsGates + properColoringGates + allColoredGates + [ccoo for ccoos in colorClassesOrderedGatesList for ccoo in ccoos] + colorabilityClassOverlappingGates + fullListAssginmentGates + subsetsGates + [admissible, outputGate]

	forallList = [item for sublist in graph.vs["listAssignment"] for item in sublist]
	forallBlock = QBlock(Quantifier.FORALL, forallList)
	existsList = [item for sublist in graph.vs["colorAssignment"] for item in sublist]
	existsBlock = QBlock(Quantifier.EXISTS, existsList)


	return QCIREncoding(FreeBlock([]), [forallBlock, existsBlock], OutputStatement("choosable"), gateList)

def basicEncoding(graph: ig.Graph, 
		lazyOptimization = True,
		overlappingOptimization=True, 
		lazyOptimizationAllColored=True, 
		lazyOptimizationOverlapping=True, 
		exactlyOneOptimization=True):
	if not lazyOptimization:
		lazyOptimizationAllColored = False
		lazyOptimizationOverlapping = False
	try:
		listsizes = graph.vs[:]["listsize"]
	except KeyError:
		print("[Error]: Graph needs listsize attribute, encoding stopped")

	variables(graph)

	lazyGates = []
	if lazyOptimization:
		lazyGates = lazyChoosable(graph)

	(inListsOut, inListsGates) = inLists(graph)
	(properColoringOut, properColoringGates) = properColoring(graph)
	(allColoredOut, allColoredGates) = allColored(graph, lazyOptimization=lazyOptimizationAllColored, exactlyOne=exactlyOneOptimization)
	(fullListAssignmentOut, fullListAssginmentGates) = fullListAssginment(graph)
	(colorClassesOrderedOut, colorClassesOrderedGates) = colorClassesOrdered(graph)
	(colorabilityClassOverlappingOut, colorabilityClassOverlappingGates) = colorabilityClassOverlapping(graph, lazyOptimization=lazyOptimizationOverlapping)

	admissible = Gate("admissible", GateConnective.AND, [Literal(True, inListsOut), Literal(True, properColoringOut), Literal(True, allColoredOut)])
	if overlappingOptimization:
		outputGate = Gate("choosable", GateConnective.OR, [Literal(False, colorClassesOrderedOut), Literal(False, colorabilityClassOverlappingOut) ,Literal(False, fullListAssignmentOut), Literal(True, "admissible")])
	else:
		outputGate = Gate("choosable", GateConnective.OR, [Literal(False, colorClassesOrderedOut), Literal(False, fullListAssignmentOut), Literal(True, "admissible")])

	gateList = lazyGates + inListsGates + properColoringGates + allColoredGates + colorClassesOrderedGates + colorabilityClassOverlappingGates + fullListAssginmentGates + [admissible, outputGate]

	forallList = [item for sublist in graph.vs["listAssignment"] for item in sublist]
	forallBlock = QBlock(Quantifier.FORALL, forallList)
	existsList = [item for sublist in graph.vs["colorAssignment"] for item in sublist]
	existsBlock = QBlock(Quantifier.EXISTS, existsList)


	return QCIREncoding(FreeBlock([]), [forallBlock, existsBlock], OutputStatement("choosable"), gateList)