#!/usr/bin/python3

import igraph as ig
import networkx as nx
import subprocess
import sys
import os
from qcir import *
import encoder as ec
import preprocessor as pp
import json
import argparse

def testGraph(graph, maxColors, maxVertices, solver="./qute", tempFilename=".temp/randomGraphTest.qcir", timeout="100s"):
	print(f"testing with {maxColors}")

	result = {"listsize": maxColors}

	(lazyChoosableGraph, lazyListSizes) = pp.lazyChoosable(graph, [maxColors]*maxVertices)
	if lazyChoosableGraph.vcount() == 0:
		result["solver_used"] = False
		result["choosable"] = True
		print("can be colored since preprocessing yields emty graph")
		return result

	(cliquesPrunedGraph, prunedListSizes, coloringPossible) = pp.cliquesOfTwins(lazyChoosableGraph, lazyListSizes, 10)

	if not coloringPossible:
		result["solver_used"] = False
		result["choosable"] = False
		print("can not be colored due to clique being too big")
		return result
		
	if cliquesPrunedGraph.vcount() == 0:
		result["solver_used"] = False
		result["choosable"] = True
		print("can be colored since preprocessing yields emty graph")
		return result

	cliquesPrunedGraph.vs["listsize"] = prunedListSizes
	 
	with open(tempFilename, "w") as f:
		f.write(str(ec.basicEncoding(cliquesPrunedGraph, lazyOptimizationAllColored=True, lazyOptimizationOverlapping=False)))
		finishedPorocess = subprocess.run(["timeout", timeout, "/usr/bin/time", "-f", '{"elapsed":"%E","userCPU_Seconds":"%U"}',  solver, tempFilename], capture_output=True)
		print(finishedPorocess.stdout)
		print(finishedPorocess.stderr)
		result["solver_used"] = True
		
		#for qute
		#if finishedPorocess.stdout == b'SAT\n':
		#	result["choosable"] = True
		#	print(str(finishedPorocess.stderr).split("\\n"))
		#	result["time"] = json.loads(str(finishedPorocess.stderr).split("\\n")[1])
		#elif finishedPorocess.stdout == b'UNSAT\n':
		#	result["choosable"] = False
		#	result["time"] = json.loads(str(finishedPorocess.stderr).split("\\n")[1])
		#elif finishedPorocess.stdout == b'UNDEF\n':
		#	result["choosable"] = False
		#	result["undecided"] = True
		
		if str(finishedPorocess.stdout).find("cnf 1"):
			result["choosable"] = True
			print(str(finishedPorocess.stderr).split("\\n"))
			result["time"] = json.loads(str(finishedPorocess.stderr).split("\\n")[2])
		elif finishedPorocess.stdout.find("cnf 0"):
			result["choosable"] = False
			result["time"] = json.loads(str(finishedPorocess.stderr).split("\\n")[2])
		else:
			result["choosable"] = False
			result["undecided"] = True
 
	os.remove(tempFilename)
	return result

def queryResults(results, solverUsed=None, minTimeUsed=None, undef=None):

	if solverUsed != None:
		results = filter(lambda result : result["solver_used"] == solverUsed, results)

	if minTimeUsed != None:
		results = filter(lambda result : result.get("time") != None and float(result["time"]["userCPU_Seconds"]) >= minTimeUsed, results)

	if undef == True:
		results = filter(lambda result : result.get("undecided") != None and result["undecided"] == True, results)

	return list(results)
	

def main():
	argParser = argparse.ArgumentParser()
	argParser.add_argument("-n", action="store")
	parsedArgs = vars(argParser.parse_args(sys.argv[1:]))
	maxVertices = int(parsedArgs["n"])

	if maxVertices == None:
		maxVertices = 4

	path = f"./.temp/all{maxVertices}VertexGraphs_cqesto"
	if not os.path.exists(path):
		os.makedirs(path)
	tempfile = f"{path}/graphs"
	if not os.path.exists(tempfile):
		subprocess.run(["./geng", "-cg", f"{maxVertices}", tempfile])

	with open(tempfile) as f:
		resultList = []
		for count, line in enumerate(f):
			graphfile = f"{path}/graph_{count}"

			if not os.path.exists(graphfile):
				print(f"checking graph no {count}")
				graph = nx.from_graph6_bytes(bytes(line.replace("\n", ""), "utf-8"))
				print(graph.edges)
				results = {"graph_edgelist": list(graph.edges), "results":[]}
				for maxColors in range(2, maxVertices):
					localResluts = testGraph(ig.Graph(graph.edges), maxColors, maxVertices, solver="./cqesto")
					results["results"].append(localResluts)

					if localResluts["choosable"] == True:
						break
					
				with open(graphfile, "w") as f:
					f.write(json.dumps(results))
				
				resultList.append(results)
			
			else:
				with open(graphfile, "r") as f:
					resultList.append(json.load(f))



	
	print("total Results")
	print(len(resultList))
	print("total Results with solver used")
	print(len([queried for results in resultList for queried in queryResults(results["results"], solverUsed=True) if queried != []]))
	print("total Results with t > 10")
	print(len([queried for results in resultList for queried in queryResults(results["results"], solverUsed=True, minTimeUsed=10) if queried != []]))
	print("total Results with timeout")
	print(len([queried for results in resultList for queried in queryResults(results["results"], undef=True) if queried != []]))

	#solverResults = []
	#for results in resultList:
	#	for result in queryResults(results["results"], solverUsed=True, minTimeUsed=10):
	#		if result["solver_used"]:
	#			solverResults.append(results)
	#			print(results)
	#			break


if __name__ == "__main__":
	main()