#!/usr/bin/python3

import igraph as ig
import timeit
from copy import deepcopy, copy
import qcir
import encoder as ec
import subprocess
from testcases import square
import os
import json
import time

class GraphPolynom:

	#n is the number of vertices, so also a hard cap on the maximum degree on one position
	def __init__(self, n):
		self.n = n
		self.coefficient = 0
		self.children = []

	def deepcopy(self):
		copy = GraphPolynom(self.n)
		copy.coefficient = self.coefficient
		for child in self.children:
			copy.children.append(child.deepcopy())

		return copy
	
	def deepcopyNeg(self):
		copy = GraphPolynom(self.n)
		copy.coefficient = -self.coefficient

		for child in self.children:
			if child.children == []:
				copy.children.append(GraphPolynom(self.n))
			else:
				copy.children.append(child.deepcopyNeg())

		return copy

	#assumes both have the same n
	def addToSelf(self, other):
		self.coefficient += other.coefficient
		if self.children == []:
			self.children = other.children
			return
		elif other.children == []:
			return
		
		for (child1, child2) in zip(self.children, other.children):
			child1.addToSelf(child2)

	def multiplyWith(self, mononomial):
		if mononomial == []:
			return
		
		if self.children == []:
			return

		self.children = [GraphPolynom(self.n) for _ in range(mononomial[0])] + self.children
		self.children = self.children[0:self.n]

		for child in self.children:
			child.multiplyWith(mononomial[1:])
	
	def pruneExponentsFrom(self, cutoffExponents):
		if cutoffExponents == []:
			return
		if self.children == []:
			return 
		
		self.children = self.children[0:cutoffExponents[0]] + [GraphPolynom(self.n) for _ in range(cutoffExponents[0], self.n)]
		for child in self.children:
			child.pruneExponentsFrom(cutoffExponents[1:])

	def insertOneValue(self, mononomial, value):
		if self.children == []:
			self.children = [GraphPolynom(self.n) for _ in range(self.n)]
		
		if mononomial == []:
			self.coefficient += value
		else:
			self.children[mononomial[0]].insertOneValue(mononomial[1:], value)


	def toList(self):
		list = []
		if self.coefficient != 0:
			list.append((self.coefficient, []))

		for i,child in enumerate(self.children):
			childList = child.toList()

			list += [(childCoefficient, [i] + childMononomial) for (childCoefficient, childMononomial) in childList]
		
		return list

	def __str__(self):
		ret = ""

		for (coefficient, mononomial) in self.toList():
			ret += f"{coefficient} * {mononomial}; "

		return ret

def oneEdgePolynom(n, edge):
	(start, end) = edge
	
	l1 = [0] * n
	l1[start] = 1

	l2 = [0] * n
	l2[end] = 1

	pol = GraphPolynom(n)

	pol.insertOneValue(l1, 1)
	pol.insertOneValue(l2, -1)

	return pol

def graphListPolynomNotEmpty(n, edgelist, maxExponents, polynom=None, previousStart=0):
	if edgelist == []:
		polynom.pruneExponentsFrom(maxExponents)
		return not polynom.toList() == []

	if polynom == None:
		polynom = oneEdgePolynom(n, edgelist[0])
		(start, _) = edgelist[0]
		return graphListPolynomNotEmpty(n, edgelist[1:], maxExponents, polynom, start)


	edgelist = sorted(edgelist, key=lambda x: x[0])

	(start, end) = edgelist[0]

	if not previousStart == start:
		for i in reversed(range(0, maxExponents[start]+1)):
			polList = [(coeff, mon) for (coeff, mon) in polynom.toList() if mon[previousStart] == i]
			
			newPol = GraphPolynom(n)
			for (coeff, mon) in polList:
				newPol.insertOneValue(mon, coeff)

			if graphListPolynomNotEmpty(n, edgelist, maxExponents, newPol, start):
				return True

		return False

	l1 = [0] * n
	l1[start] = 1

	l2 = [0] * n
	l2[end] = 1

	secondPolynom = polynom.deepcopyNeg()
	
	polynom.multiplyWith(l1)
	secondPolynom.multiplyWith(l2)

	polynom.addToSelf(secondPolynom)
	polynom.pruneExponentsFrom(maxExponents)

	return graphListPolynomNotEmpty(n, edgelist[1:], maxExponents, polynom, start)

def graphListPolynom(n, edgelist, maxExponents=None):

	polynom = oneEdgePolynom(n, edgelist[0])

	if maxExponents:
		polynom.pruneExponentsFrom(maxExponents)

	for i,(start, end) in enumerate(edgelist[1:], 1):
		secondPolynom = polynom.deepcopyNeg()

		l1 = [0] * n
		l1[start] = 1

		l2 = [0] * n
		l2[end] = 1

		polynom.multiplyWith(l1)
		secondPolynom.multiplyWith(l2)

		polynom.addToSelf(secondPolynom)

		if maxExponents:
			polynom.pruneExponentsFrom(maxExponents)

	return polynom

def allSubsets(start, end):
	if start >= end:
		yield []
	else:
		for subset in allSubsets(start+1, end):
			yield [start] + subset
			yield subset

#there is some bug in the indexing
def checkIfReducible(graph, subsets):
	if subsets == []:
		edgelist = [(e.source, e.target) for e in graph.es]
		if edgelist != []:
			if graphListPolynomNotEmpty(len(graph.vs), edgelist, graph.vs[:]["listsize"]):
				return True
			else:
				return False
		else:
			for listsize in graph.vs[:]["listsize"]:
				if listsize == 0:
					return False
			return True
	

	if subsets[0] == []:
		return checkIfReducible(graph, subsets[1:])

	subgraph = graph.subgraph(subsets[0])

	for independentSet in subgraph.maximal_independent_vertex_sets():

		independentSetOriginalIndices = []
		for index in independentSet:
			independentSetOriginalIndices.append(subgraph.vs[index]["originalIndex"])

		complementIndices = []
		for i in range(graph.vcount()):
			if not i in independentSetOriginalIndices:
				complementIndices.append(i)

		copygraph = deepcopy(graph)
		
		newListSizes = []
		for i, listsize in enumerate(graph.vs[:]["listsize"]):
			if i in subsets[0]:
				newListSizes.append(listsize - 1)
			else:
				newListSizes.append(listsize)

		copygraph.vs[:]["listsize"] = newListSizes
		complementGraph = copygraph.subgraph(complementIndices)
		for i,v in enumerate(complementGraph.vs):
			v["originalIndex"] = i

		newSubsets = []

		for subset in subsets[1:]:
			newSubset = [v for v in subset if v not in independentSetOriginalIndices]
			for isoIndex in reversed(independentSetOriginalIndices):
				for i,v in enumerate(newSubset):
					if v > isoIndex:
						newSubset[i] -= 1
			newSubsets.append(newSubset)


		if checkIfReducible(complementGraph, newSubsets):
			return True
	return False

def checkChoosabilityClasses(graph, selected, possible, minPossible, reducibleThreshold=8, timesCalled=0, maximumTimesCalled=2**30):
	timesCalled += 1
	if timesCalled % 2**10 == 0:
		print(f"times called: {timesCalled}")
	if timesCalled > maximumTimesCalled:
		print("cutoff - maximum times called exceeded")
		return (False, -1)
	listsizes = [0] * graph.vcount()
	for colorabilityClass in selected:
		for vertexIndex in colorabilityClass:
			listsizes[vertexIndex] += 1

	
	full = True
	for (vertex, listsize) in zip(graph.vs, listsizes):
		if vertex["listsize"] < listsize:
			return (True, timesCalled)
		elif vertex["listsize"] != listsize:
			full = False
	
	if full:
		reducible =  checkIfReducible(graph, selected)
		#if not reducible: print(f"full la: {selected} is not reducible")
		return (reducible, timesCalled)


	if len(selected) > 0 and len(selected) < reducibleThreshold and checkIfReducible(graph, selected):
		#print("reduciblity found")
		return (True, timesCalled)
	else:
		for i,colorabilityClass in enumerate(possible[minPossible:], minPossible):
			if colorabilityClass != []:
				(check, timesCalled) = checkChoosabilityClasses(graph, selected + [colorabilityClass], possible, i, timesCalled=timesCalled, maximumTimesCalled=maximumTimesCalled)
				if not check:
					return (False, timesCalled)
		return (True, timesCalled)

#graph = ig.Graph([(0,1),(0,2),(1,2)])
#graph.vs[:]["listsize"] = [1,1,3]


#these graphs come from the suite of all 8 vertex graphs that were checked
#testgraph = ig.Graph([(0, 5), (0, 6), (0, 7), (1, 5), (1, 6), (1, 7), (2, 5), (2, 6), (2, 7), (3, 5), (3, 6), (3, 7), (4, 7)])
#testgraph = ig.Graph([(0, 3), (0, 4), (0, 6), (0, 7), (1, 3), (1, 5), (1, 6), (2, 4), (2, 5), (3, 5), (3, 7), (4, 6), (4, 7), (5, 7)])
#testgraph.vs[:]["listsize"] = [3] * 8

#testgraph.vs[:]["originalIndex"] = range(0,8)


#testgraph = ig.Graph([(0,3),(0,4),(0,5),(1,3),(1,4),(1,5),(2,3),(2,4),(2,5)])
#testgraph.vs[:]["listsize"] = [2]*6
#testgraph.vs[:]["originalIndex"] = range(0,6)

def calcNonReduicbleSubsets(testgraph, verbose=0, returnRecursiveCalls=False):
	reducibleSetsCount = 0
	reducibleSets = []
	reducibleSetsMap = []
	for subset in allSubsets(0,testgraph.vcount()):
		subgraph = testgraph.induced_subgraph(subset)
		if verbose>0:print("subset:")
		if verbose>0:print(subset)
		if verbose>0:print("vetting subgraph:")
		if verbose>0:print(subgraph)
		
		previousSetFound = False
		for (previousSet, previousIndependentSet) in reducibleSetsMap:
			if set(subset) <= set(previousSet) and set(subset) >= set(previousIndependentSet):
				previousSetFound = True
				if verbose>0:print(f"found greater subset {previousSet} with smaller independentSet {previousIndependentSet}")
				reducibleSets.append(subset)
				reducibleSetsCount += 1
				break
			
		if previousSetFound:
			continue
		
		for independentSet in subgraph.maximal_independent_vertex_sets():
			if verbose>0:print("independentSet")
			if verbose>0:print(list(independentSet))
	
			independentSetOriginalIndices = []
			for index in independentSet:
				independentSetOriginalIndices.append(subgraph.vs[index]["originalIndex"])
	
			complementIndices = []
			for i in range(testgraph.vcount()):
				if not i in independentSetOriginalIndices:
					complementIndices.append(i)
	
			copygraph = deepcopy(testgraph)
			
			newListSizes = []
			for i, listsize in enumerate(testgraph.vs[:]["listsize"]):
				if i in subset:
					newListSizes.append(listsize - 1)
				else:
					newListSizes.append(listsize)
	
			copygraph.vs[:]["listsize"] = newListSizes
			complementGraph = copygraph.subgraph(complementIndices)
	
	
			complementEdges = [(e.source, e.target) for e in list(complementGraph.es)]
	
			if complementEdges != []:
				if graphListPolynomNotEmpty(len(complementGraph.vs), complementEdges, complementGraph.vs[:]["listsize"]):
					reducibleSets.append(subset)
					reducibleSetsMap.append((subset, independentSetOriginalIndices))
					reducibleSetsCount += 1
					break
				else:
					if verbose>0:print("looking for next independent set")
			else: 
				reducibleSets.append(subset)
				reducibleSetsMap.append((subset, independentSetOriginalIndices))
				reducibleSetsCount += 1
				break
	
	if verbose > 0: print(f"length of reducible subsets map: {len(reducibleSetsMap)}")
	if verbose > 0: print(f"num reducible Subsets: {reducibleSetsCount}")

	nonReducibleSets = []
	for subset in allSubsets(0,testgraph.vcount()):
		if not subset in reducibleSets:
			nonReducibleSets.append(subset)
	
	return nonReducibleSets

def testIfChoosable(testgraph, verbose=0, returnRecursiveCalls=False, returnNonReducibleSubsets=False):
	nonReducibleSets = calcNonReduicbleSubsets(testgraph, verbose)

	(reducible, recursiveCalls) = checkChoosabilityClasses(testgraph, [], nonReducibleSets, 0, timesCalled=0, maximumTimesCalled=2**17)
	
	if verbose > 0: print("reducible:")
	if verbose > 0: print(reducible)
	if verbose > 0: print(f"recursive checkChoosabilityClasses calls needed: {recursiveCalls}")
	if returnNonReducibleSubsets and returnRecursiveCalls:
		return (reducible, recursiveCalls, nonReducibleSets)
	if returnRecursiveCalls:
		return (reducible, recursiveCalls)
	if returnNonReducibleSubsets:
		return (reducible, nonReducibleSets)
	return reducible

def findGraphVettingBad():
	n = 8
	writeIndex = 0
	while True:
		graph = ig.Graph.K_Regular(n, 4)
		f = [3]*n
		graph.vs[:]["listsize"] = f
		graph.vs[:]["originalIndex"] = range(0,n)

		choosable = testIfChoosable(graph, 0)
		
		repeat = choosable
		while repeat:
			repeat = False
			for i in range(n):
				f[i] -= 1
				graph.vs[:]["listsize"] = f

				(choosable, calls, nonReducibleSubsets) = testIfChoosable(graph, verbose=0, returnRecursiveCalls=True, returnNonReducibleSubsets=True)

				result = {
					"graph": list(map(lambda x: x.tuple, graph.es)),
					"listSizeFunction": f,
					"nonReducibleSubsets": nonReducibleSubsets,
					"recursiveCalls": calls,
					"foundToBeChoosable": choosable
				}

				print(f"writing with index {writeIndex}")
				with open(f".temp/graphsVettingMethod{n}Vertices/result_{writeIndex}_subsets_{nonReducibleSubsets}.json", "w") as file:
					file.write(json.dumps(result))
				writeIndex += 1


				if not choosable:
					f[i] += 1
				else:
					repeat = True



def main():
	findGraphVettingBad()
	graphVettingBadList = [
		(0,2),(0,5),(0,6),(0,8),
		(1,3),(1,4),(1,7),(1,8),
		(2,3),(2,4),(2,5),
		(3,6),(3,7),
		(4,6),(4,9),
		(5,7),(5,8),
		(6,9),
		(7,9),
		(8,9)]
	graphVettingBad = ig.Graph(graphVettingBadList)
	f = [2, 2, 3, 3, 2, 3, 3, 3, 3, 3]
	graphVettingBad.vs[:]["listsize"] = f
	graphVettingBad.vs[:]["originalIndex"] = range(0,graphVettingBad.vcount())
	testIfChoosable(graphVettingBad, verbose=1)

#	graphFromNielsenList =  [(0,1),(0,2),(1,2),(1,3),(3,4),(4,5),(4,6),(6,7),(7,8),(5,8),(5,9),(9,10),(9,11),(10,11)]
#	f = [3,4,3,4,5,6,3,2,3,5,3,3]
#	testgraph = ig.Graph(square(graphFromNielsenList))
#	testgraph.vs[:]["listsize"] = f
#	testgraph.vs[:]["originalIndex"] = range(0,12)
#	print(testgraph.vcount())
#	testIfChoosable(testgraph, verbose = 1)

#	graphFromNielsenList =  [(0,1),(0,2),(1,2),(1,3),(3,4),(4,5),(5,6),(5,7),(6,7)]
#	f = [3,4,3,3,3,4,3,3]
#	testgraph = ig.Graph(square(graphFromNielsenList))
#	testgraph.vs[:]["listsize"] = f
#	testgraph.vs[:]["originalIndex"] = range(0,8)

#	testIfChoosable(testgraph, verbose = 1)

if __name__ == "__main__":
	main()